#include "spmd_internal.h"
#include "task.h"
#include "atomic.h"
#include <stdlib.h>


void spmd_spawn(int nthreads, void (*func)(void *), void *arg) 
{
  assert(nthreads > 0);
  SPMD_Scheduler *sched = new SPMD_Scheduler(nthreads, func, arg);
  sched->spawn();
  delete sched;
}


int spmd_tid() 
{
  void *tid;
  lithe_task_gettls(&tid);
  return (int)(long)tid;
}


SPMD_Scheduler::SPMD_Scheduler(int __nthreads, void (*__func)(void *), void *__arg) : 
  nthreads(__nthreads),
  func(__func),
  arg(__arg),
  spawned(0),
  completed(0)
{
  spinlock_init(&unreg_once);
  unblocked = new TaskQueue(__nthreads);
}


SPMD_Scheduler::~SPMD_Scheduler() 
{
  delete unblocked;
}


void SPMD_Scheduler::spawn() 
{
  lithe_sched_register(&funcs, (void *)this, __spawn, (void *)this);
}


void SPMD_Scheduler::enter() 
{
  lithe_task_t *task;

  /* nothing left to do: release hart*/
  if (coherentread(&completed) == nthreads) {
    if (spinlock_trylock(&unreg_once) == UNLOCKED) {
      lithe_sched_unregister();
    } else {
      lithe_sched_yield();
    }
  }

  /* complete unblocked contexts */
  task = unblocked->pop();
  if (task != NULL) {
    lithe_task_resume(task);
  }

  /* spawn unstarted threads on a new context */
  task = task_create();
  lithe_task_do(task, __enter_begin, (void *)this);
}


void SPMD_Scheduler::unblock(lithe_task_t *task) 
{
  unblocked->push(task);
  lithe_sched_request(1);
}


void SPMD_Scheduler::enter_begin() 
{
  bool done;

  /* spawn next thread */
  int tid = fetchadd(&spawned, 1);
  while (tid < nthreads) {
    lithe_task_settls((void *)tid);
    func(arg);
    atomic_increment(&completed);
    tid = fetchadd(&spawned, 1);
  }

  /* completely done: clean up context & unregister scheduler */
  if ((coherentread(&completed) == nthreads) && (spinlock_trylock(&unreg_once) == UNLOCKED)) {
    lithe_task_block(__enter_end, (void *)ACTION_UNREG);
  } 
  
  /* clean up current context & resume an unblocked context */
  lithe_task_t *task = unblocked->pop();
  if (task != NULL) {
    lithe_task_block(__enter_end, (void *)task);
  }
    
  /* nothing left to do (for now): clean up context and release hart */
  lithe_task_block(__enter_end, (void *)ACTION_YIELD);
}


void __spawn(void *__sched) 
{
  lithe_sched_request(((SPMD_Scheduler *)__sched)->nthreads);
  ((SPMD_Scheduler *)__sched)->enter();
}


void __enter_begin(void *__sched) 
{
  ((SPMD_Scheduler *)__sched)->enter_begin();
}


void __enter_end(lithe_task_t *task, void *__action) 
{
  task_destroy(task);
  long action = (long)__action;
  if (action == ACTION_YIELD) {
    lithe_sched_yield();
    assert(0);
  } else if (action == ACTION_UNREG) {
    lithe_sched_unregister();
    assert(0);
  } else {
    lithe_task_resume((lithe_task_t *)__action);
    assert(0);
  }
}






  
  



